from __future__ import annotations

from .._git import get_commits
from ._base import Command


class History(Command):
    """Show how the baseline changed over time.
    """

    def run(self) -> int:
        prev_count: int | None = None
        self.print('date       time           res   old  fix  new   commit      author')
        for commit in get_commits(self.config.baseline_path):
            commit.fix_lines_count(prev_count)
            count_formatted = f'{commit.lines_count:>3}'
            line = f'{commit.created_at} {self.colors.blue(count_formatted)}'
            line += f' = {prev_count or 0:>3}'
            if commit.deletions:
                formatted = f'{-commit.deletions: >+4}'
                line += f' {self.colors.green(formatted)}'
            else:
                line += ' ' * 5
            if commit.insertions:
                formatted = f'{commit.insertions: >+4}'
                line += f' {self.colors.red(formatted)}'
            else:
                line += ' ' * 5
            line += f'   {commit.hash[:10]}'
            line += f'  {self.colors.magenta(commit.author_email)}'
            self.print(line)
            prev_count = commit.lines_count
        return 0



from __future__ import annotations

from argparse import ArgumentParser
from collections import defaultdict

from .._error import Error
from ._base import Command


class TopFiles(Command):
    """Show files with the most errors.
    """
    @classmethod
    def init_parser(cls, parser: ArgumentParser) -> None:
        super().init_parser(parser)
        parser.add_argument(
            '-n', type=int, default=20,
            help='how many files to show',
        )

    def run(self) -> int:
        baseline_text = self.config.baseline_path.read_text(encoding='utf8')
        file_stats: dict[str, int] = defaultdict(int)
        baseline = baseline_text.strip().splitlines()
        for line in baseline:
            error = Error.new(line)
            if error is None:
                self.print(f'cannot parse line: {line}')
                return 1
            path = '/'.join(error.path.parts[:self.config.depth])
            file_stats[path] += 1

        sorted_file_stats = sorted(
            file_stats.items(),
            key=lambda x: x[::-1],
            reverse=True,
        )
        max_width = max(len(path) for path, _ in sorted_file_stats[:self.args.n])
        for path, total_count in sorted_file_stats[:self.args.n]:
            total_formatted = f'{total_count:>3}'
            path = path.ljust(max_width)
            percent = round(total_count / len(baseline) * 100, 1)
            self.print(f'{path} {self.colors.blue(total_formatted)} {percent:>4}%')

        return 0



from __future__ import annotations

from argparse import ArgumentParser, Namespace
from dataclasses import dataclass
from functools import cached_property
from typing import ClassVar, TextIO

from .._colors import Colors
from .._config import Config


@dataclass
class Command:
    name: ClassVar[str]
    args: Namespace
    stdin: TextIO
    stdout: TextIO

    @staticmethod
    def init_parser(parser: ArgumentParser) -> None:
        Config.init_parser(parser)

    def run(self) -> int:
        raise NotImplementedError

    def print(self, *args: str, end: str = '\n', sep: str = ' ') -> None:
        print(*args, file=self.stdout, end=end, sep=sep)

    @cached_property
    def config(self) -> Config:
        return Config.from_args(vars(self.args))

    @cached_property
    def colors(self) -> Colors:
        return Colors(disabled=self.config.no_colors)



from __future__ import annotations

import os
import subprocess
from argparse import ArgumentParser
from contextlib import contextmanager
from io import StringIO
from pathlib import Path
from typing import Iterator

import pytest

from mypy_baseline._config import Config
from mypy_baseline.commands import Suggest

from .helpers import LINE1, run


@contextmanager
def enter_dir(path: Path) -> Iterator[None]:
    old_path = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(old_path)


@contextmanager
def set_env_vars(env_vars: dict[str, str]) -> Iterator[None]:
    old = os.environ.copy()
    try:
        os.environ.update(env_vars)
        yield
    finally:
        os.environ.clear()
        os.environ.update(old)


def run_git(*args):
    subprocess.run(['git', *args]).check_returncode()


@pytest.fixture
def repo_path(tmp_path: Path) -> Iterator[Path]:
    repo_path = tmp_path / 'repo'
    repo_path.mkdir()
    with enter_dir(repo_path):
        run_git('init')
        yield repo_path


def test_suggest(repo_path: Path):
    bline_path = repo_path / 'baseline.txt'
    bline_path.write_text(LINE1)
    run_git('add', str(bline_path))
    run_git('commit', '-m', 'init')
    run_git('checkout', '-b', 'feature-branch')
    cmd = ['suggest', '--baseline-path', str(bline_path)]
    stdout = run(cmd, exit_code=1)
    assert LINE1.strip() == stdout.strip()


def test_target_branch__cli_flag():
    parser = ArgumentParser()
    Config.init_parser(parser)
    args = parser.parse_args(['--default-branch', 'something'])
    cmd = Suggest(args=args, stdin=StringIO(), stdout=StringIO())
    assert cmd.target == 'something'


@pytest.mark.parametrize('env_var', [
    'CI_MERGE_REQUEST_TARGET_BRANCH_SHA',
    'GITHUB_BASE_REF',
])
def test_target_branch__env_var(env_var: str):
    parser = ArgumentParser()
    Config.init_parser(parser)
    args = parser.parse_args([])
    with set_env_vars({env_var: 'something'}):
        cmd = Suggest(args=args, stdin=StringIO(), stdout=StringIO())
        assert cmd.target == 'something'
